/***************************************************************************\
  "icm.hpp" - Intermediate Code Manager (Phase D)
  Revision 14; 10 Nov 2002 - 15 March 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

#ifndef ICM_HPP
#define ICM_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

class IntermediateCodeManager
{
  public:
    IntermediateCodeManager ();
    ~IntermediateCodeManager ();

    int replace (int x, Statement *statement);
    int put (Statement *statement);
    void putLine ();
    void reset ();
    void goToPosition (int x);
    Statement *getNext ();
    Statement *get (int x);

    int xLast; // index of last statement in the mvProgram vector
    int xCurr; // index of current statement in the mvProgram vector

  private:
    Statement *mvProgram[GLOBAL_ICM_MAX_INSTRUCTIONS]; // vector of statements
    bool mReset;
};

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

IntermediateCodeManager::IntermediateCodeManager ()
{
  for (int i = 0; i < GLOBAL_ICM_MAX_INSTRUCTIONS; i++)
  {
    mvProgram[i] = NULL;
  }

  xLast = 0;
  xCurr = 0;
  mReset = false;
}

IntermediateCodeManager::~IntermediateCodeManager ()
{
  for (int i = 0; i < GLOBAL_ICM_MAX_INSTRUCTIONS; i++)
  {
    if (mvProgram[i]) delete mvProgram[i];
  }
}

//// Replace an existing statement with another one. There must
////   already be a statement in the existing position.
int IntermediateCodeManager::replace (int x, Statement *statement)
{
  if (mReset == true)
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    errterm ();
  }

  if (mvProgram[x])
  {
    delete mvProgram[x];
    mvProgram[x] = statement;
  }
  else
  {
    err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, "Tried to replace non-existant statement");
    errterm ();
  }

  return x;
}

int IntermediateCodeManager::put (Statement *statement)
{
  if (mReset == false)
  {
    if ((xCurr + 1) > GLOBAL_ICM_MAX_INSTRUCTIONS)
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      errterm ();
    }

    mvProgram[xCurr] = statement;
    xCurr++;
    xLast++;
  }
  else
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    errterm ();
  }

  return (xCurr - 1);
}

void IntermediateCodeManager::putLine ()
{
  if (mReset == false)
  {
    Symbol *symbol;

    if ((xCurr + 1) > GLOBAL_ICM_MAX_INSTRUCTIONS)
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      errterm ();
    }

    if (Global::bDebug)
    {
      // Insert DBGFILE instruction
      symbol = new Symbol (Global::sCurrentFilename);
      symbol->set (repLiteral, tStr, false, Global::sCurrentFilename);
      gScopeManager->addSymbolToList (symbol); // add symbol to deletion list

      mvProgram[xCurr] = new Statement (ttDBGFILE, symbol);
      xCurr++;
      xLast++;

      // Insert DBGLINE instruction
      symbol = new Symbol ("line_number");
      symbol->set (repLiteral, tInt, false, &Global::nCurrentLine);
      gScopeManager->addSymbolToList (symbol); // add symbol to deletion list

      mvProgram[xCurr] = new Statement (ttDBGLINE, symbol);
      xCurr++;
      xLast++;
    }

    if (Global::bPrintLine)
    {
      // Insert __Buffer instruction
      symbol = new Symbol (Global::sCurrentFilename);
      char *sTemp = k_trim(gInputManager->mBufferStackTop->text);
      symbol->set (repLiteral, tStr, false, sTemp);
      k_del_s (sTemp);
      gScopeManager->addSymbolToList (symbol); // add symbol to deletion list

      mvProgram[xCurr] = new Statement (tt__Buffer, symbol);
      xCurr++;
      xLast++;
    }
  }
  else
  {
    err (true, EI_WRITE_AFTER_RESET, NULL);
    errterm ();
  }
}

void IntermediateCodeManager::reset ()
{
  xCurr = 0;
  mReset = true;
}

void IntermediateCodeManager::goToPosition (int x)
{
  if (mReset == true)
  {
    if ((x > xLast) || (x > GLOBAL_ICM_MAX_INSTRUCTIONS))
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      errterm ();
    }

    xCurr = x;
  }
  else
  {
    err (true, EI_GOTO_BEFORE_RESET, NULL);
    errterm ();
  }
}

Statement *IntermediateCodeManager::getNext ()
{
  if (mReset == true)
  {
    if ((xCurr + 1) > xLast)
    {
      return NULL; // there are no more statements, so return NULL
    }

    return mvProgram[xCurr++];
  }
  else
  {
    err (true, EI_READ_BEFORE_RESET, NULL);
    errterm ();
    return NULL;
  }
}

Statement *IntermediateCodeManager::get (int x)
{
  if (mReset == true)
  {
    if ((x > xLast) || (x > GLOBAL_ICM_MAX_INSTRUCTIONS))
    {
      err (true, EI_CURR_INDEX_OUT_OF_BOUNDS, NULL);
      errterm();
    }

    return mvProgram[x];
  }
  else
  {
    err (true, EI_READ_BEFORE_RESET, NULL);
    errterm();
    return NULL;
  }
}

#endif
