/***************************************************************************\
  "k_string.hpp" - Kevin's String Manipulation Functions ( C++ )
  Revision 76; 3 June 2000 - 15 March 2004
  Available at: "http://kevin.alteu.com/"
  E-mail:       "kevin@albrecht.net"

  Copyright (C) 2000-2004, Kevin P. Albrecht

  Permission is hereby granted by Kevin P. Albrecht (the "Author"), free of
  charge, to any person obtaining a copy of this software and associated
  documentation files (the "Software"), to deal in the Software without
  restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the
  Software, and to permit persons to whom the Software is furnished to do
  so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included
  in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
  THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
  AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

  Except as contained in this notice, the name of the Author shall not be
  used in advertising or otherwise to promote the sale, use or other
  dealings in this Software without prior written authorization from the
  Author.
\***************************************************************************/

#ifndef K_STRING_HPP
#define K_STRING_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

// External Defines //

// External Usage //
#define K_NUMERICS "0123456789-."
#define K_DIGITS "0123456789"
#define K_LETTERS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define K_SPECIAL1 "~`!@#$%^&*()_-+={[}]|\\:;\"'<,>.?/"
#define K_SPECIAL2 "*()-+=:;\"<,>./&"
#define K_LOWERCASE "abcdefghijklmnopqrstuvwxyz"
#define K_UPPERCASE "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define K_WHITESPACE "\t\0\n "

// Internal Usage Only //
#define K_STRING_K_SWAP_STRING_SIZE 256
#define K_STRING_K_CONV_DS_STR_SIZE 256
#define K_STRING_K_CONV_IS_STR_SIZE 256
#define K_STRING_K_POSS_TEMP_SIZE 256

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //
#include <cstring>
#include <cstdlib>
#include <cstdio>
using namespace std;

// Local //

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

char *k_conv_bs (bool ab);
char *k_conv_is (int n);
char *k_conv_fs (float f, int places);
char *k_conv_ds (double d, int places);
char *k_conv_cs (char ac);
bool k_conv_sb (char *s);
bool k_conv_sbD (char *aS);
int k_conv_si (char *aS);
int k_conv_siD (char *aS);
float k_conv_sf (char *aS);
float k_conv_sfD (char *aS);
double k_conv_sd (char *aS);
double k_conv_sdD (char *aS);
float *k_new_f (float f);
double *k_new_d (double d);
bool *k_new_b (bool b);
int *k_new_n (int n);
char *k_new_s (int size);

char *k_app (char *one, char *two);
void k_cat (char *one, char *two);
void k_catd (char *one, char *two);
char k_char (char *str, int pos);
char *k_clone (char *as);
void k_copy (char *one, char *two);
void k_copyd (char *one, char *two);
void k_del_s (char *str);
void k_error (bool terminate, char *s);
bool k_setfind (char ch, char *str);
int k_posc (char *str, char ch, int after);
int k_poss (char *str, char *sub, int after);
bool k_same (char *one, char *two);
char *k_sub (char *str, int one, int two);
char *k_subd (char *str, int one, int two);
char *k_swap (char *str, char *one, char *two);
char *k_trim (char *str);
char *k_trimd (char *str);

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

char *k_conv_bs (bool ab)
{
  if (ab) return "true";
  else    return "false";
}

//// Returns the string equivalent of the integer ////
char *k_conv_is (int n)
{
  char *s = k_new_s (K_STRING_K_CONV_IS_STR_SIZE);
  sprintf (s, "%d", n);

  /* ITOA is not C Standard 
   * k_copy (s, "");
   * itoa (n, s, 10);
   */

  return s;
}

//// Returns the string equivalent of the float ////
char *k_conv_fs (float f, int places)
{
  return k_conv_ds ((double) f, places);
}

//// Returns the string equivalent of the double ////
char *k_conv_ds (double d, int places)
{
  int dec = 0;      // the location of the decimal point
  int sign = 0;     // the sign of the double value
  int i = 0;        // iterator
  char *s = NULL;   // string copied from the double
  char *buf = NULL; // holds string as it is modified

  if (d == 0) return k_clone("0");

  s = (char *) fcvt (d, places, &dec, &sign);

  buf = k_new_s (K_STRING_K_CONV_DS_STR_SIZE);
  if (sign != 0) k_copy (buf, "-");
  else           k_copy (buf, "");

  if (dec <= 0)
  {
    k_cat (buf, ".");
    while (dec < 0)
    {
      k_cat (buf, "0");
      dec++;
    }
    k_cat (buf, s);
  }
  else
  {
    k_catd (buf, k_sub (s, 0, dec));
    k_cat (buf, ".");
    k_catd (buf, k_sub (s, dec, strlen (s)));
  }

  for (i = strlen (buf) - 1; i >= 0; i--)
  {
    switch (buf[i])
    {
      case '0':
        buf[i] = '\0';
        break;
      case '.':
        buf[i] = '\0';
        goto OUT;
      default:
        goto OUT;
    }
  }
  OUT:

  return buf;
}

char *k_conv_cs (char ac)
{
  char *s = k_new_s (2);
  s[0] = ac;
  s[1] = '\0';
  return s;
}

bool k_conv_sb (char *s)
{
  if (s == NULL)
  {
    return false;
  }
  else if (k_same (s, "false"))
  {
    return false;
  }
  else if (k_conv_sd (s) == 0.0)
  {
    return false;
  }
  else
  {
    return true;
  }
}

bool k_conv_sbD (char *aS)
{
  bool temp = k_conv_sb (aS);
  k_del_s (aS);
  return temp;
}

int k_conv_si (char *aS)
{
  if (aS == NULL) return 0;
  return (atoi (aS));
}

int k_conv_siD (char *aS)
{
  int temp = k_conv_si (aS);
  k_del_s (aS);
  return temp;
}

float k_conv_sf (char *aS)
{
  if (aS == NULL) return 0.0;
  return ((float) atof (aS));
}

float k_conv_sfD (char *aS)
{
  float temp = k_conv_sf (aS);
  k_del_s (aS);
  return temp;
}

double k_conv_sd (char *aS)
{
  if (aS == NULL) return 0.0;
  return (atof (aS));
}

double k_conv_sdD (char *aS)
{
  double temp = k_conv_sd (aS);
  k_del_s (aS);
  return temp;
}

float *k_new_f (float f)
{
  float *pf = new float;
  *pf = f;
  return pf;
}

double *k_new_d (double d)
{
  double *pd;
  pd = new double;
  *pd = d;
  return pd;
}

bool *k_new_b (bool b)
{
  bool *pb;
  pb = new bool;
  *pb = b;
  return pb;
}

int *k_new_n (int n)
{
  int *pn;
  pn = new int;
  *pn = n;
  return pn;
}

//// Allocate memory for a string; if size=0, call k_error ////
char *k_new_s (int size)
{
  void *temp;

  if (size <= 0)
  {
    k_error (true /*terminate*/, "k_new_s");
  }

  temp = new char[size + 1];

  return (char*) temp;
}

//// k_app
//// Outputs the concatenation of 'one' and 'two'
//// If the inputs are NULL, they are given an empty string
char *k_app (char *one, char *two)
{
  char *temp;
  if (one == NULL) strcpy (one, "");
  if (two == NULL) strcpy (two, "");

  temp = k_new_s (strlen(one) + strlen(two));
  k_copy (temp, one);
  k_cat (temp, two);
  return temp;
}

//// k_cat
//// Concatenates 'two' to the end of 'one'
void k_cat (char *one, char *two)
{
  strcat (one, two);
}

//// k_catd
//// Concatenates 'two' to the end of 'one', then deletes 'two'
void k_catd (char *one, char *two)
{
  strcat (one, two);
  k_del_s (two);
}

//// k_char
//// Returns the char in the specified position in the specified string
char k_char (char *str, int pos)
{
  if ((pos < 0) || (pos >= (signed) strlen (str)))
  {
    // out of bounds //
    return 0;
  }
  return str[pos];
}

//// k_clone
char *k_clone (char *as)
{
  char *temp;

  if (as != NULL)
  {
    // If the following line is in the code,
    //  when "as" == "", NULL will be returned.
    /* if (as[0] == 0) return NULL; */

    temp = k_new_s (strlen(as) + 1);
    k_copy (temp, as);
    return temp;
  }
  else
  {
    return NULL;
  }
}

//// k_copy
//// Copies 'two' into 'one'
void k_copy (char *one, char *two)
{
  if (two == NULL) strcpy (one, "");
  else             strcpy (one, two);
}

//// k_copyd
//// Copies 'two' into 'one', then deletes 'two'
void k_copyd (char *one, char *two)
{
  if (two == NULL) strcpy (one, "");
  else             strcpy (one, two);
  k_del_s (two);
}

//// k_del
//// Deletes the string if it is not null or of zero len
void k_del_s (char *str)
{
  if (str == NULL) return;

  if (strlen (str) > 0)
  {
    delete [] str;
  }
}

//// k_error
//// Signal a VERY BAD error
void k_error (bool terminate, char *s)
{
  if (terminate == true)
  {
    printf ("K_ERROR: TERMINAL ERROR IN: %s\n", s);
    exit (0);
  }

  printf ("K_ERROR: RECOVERABLE ERROR IN: %s\n", s);
}

//// Returns true if 'ch' is in 'str'
bool k_setfind (char ch, char *str)
{
  if (k_posc (str, ch, 0) < 0)
    return false;
  else
    return true;
}

//// k_posc
//// Returns the position of the first occurance of the char after 'after';
//// returns -1 if not found
int k_posc (char *str, char ch, int after)
{
  int i;

  if ((after >= (signed) strlen (str)) || (after < 0))
  {
    // out of bounds //
    return -1;
  }

  for (i = after; i < (signed) strlen (str); i++)
  {
    if (str[i] == ch)
    {
      // found //
      return i;
    }
  }

  // not found //
  return -1;
}

//// k_poss
//// Returns the position of the first occurance of the string after 'after';
//// returns -1 if not found
int k_poss (char *str, char *sub, int after)
{
  int str_len, sub_len, temp_len, str_pos, sub_pos;
  char temp[K_STRING_K_POSS_TEMP_SIZE];

  if ((after >= (signed) strlen (str)) || (after < 0) || ((signed) strlen (sub) > (signed) strlen (str)))
  {
    // out of bounds //
    return -1;
  }

  // variables //
  str_len = strlen (str);
  sub_len = strlen (sub);
  temp_len = 0;

  // determine if substring is present:
  for (str_pos = after; str_pos <= str_len; str_pos++)
  {
    if (str[str_pos] == sub[0])
    {
      for (sub_pos = 0; sub_pos < sub_len; sub_pos++)
      {
        temp[sub_pos] = str[str_pos + sub_pos];
        temp_len = sub_pos;
      }
      temp[temp_len + 1] = '\0';
      if (k_same (temp, sub))
      {
        return str_pos;
      }
    }
  }

  return -1;
}

//// k_same
//// Returns true if 'one' and 'two' are the same
bool k_same (char *one, char *two)
{
  if ((one[0] == 0) || (two[0] == 0))
  {
    k_error (true /*terminate*/, "k_same");
  }

  return (strcmp (one, two) == 0);
}

//// k_sub
//// Returns the substring between two points; '0' if out of bounds
char *k_sub (char *str, int one, int two)
{
  int str_len, str_pos, sub_pos;
  char *sub;

  str_len = strlen (str);

  if ((one < 0) || (one > str_len) || (two < one))
  {
    // out of bounds //
    return 0;
  }
  if (one == two)
  {
    // zero length //
    return "";
  }

  sub = k_new_s (two - one + 1);
  strcpy (sub, "");

  // create substring //
  for (str_pos = one, sub_pos = 0; str_pos < two; str_pos++, sub_pos++)
  {
    sub[sub_pos] = str[str_pos];
  }
  sub[two - one] = '\0'; // end with '\0'

  return sub;
}

//// k_sub
//// Returns the substring between two points; '0' if out of bounds;
////  deletes the argument
char *k_subd (char *str, int one, int two)
{
  int str_len, str_pos, sub_pos;
  char *sub;

  str_len = strlen (str);
  sub = k_new_s (two - one + 1);
  strcpy (sub, "");

  if ((one < 0) || (one > str_len) || (two < one))
  {
    // out of bounds //
    k_del_s (sub);
    k_del_s (str);
    return 0;
  }
  if (one == two)
  {
    // zero length //
    k_del_s (sub);
    k_del_s (str);
    return "";
  }

  // create substring //
  for (str_pos = one, sub_pos = 0; str_pos < two; str_pos++, sub_pos++)
  {
    sub[sub_pos] = str[str_pos];
  }
  sub[two - one] = '\0'; // end with '\0'

  k_del_s (str);
  return sub;
}


//// k_swap
//// Swaps all the characters in a string that occur in the first array with matching
////  characters in the second array; returns the string with swapped characters;
////  '0' if errors
char *k_swap (char *str, char *one, char *two)
{
  int str_len, array_len, str_pos, array_pos;
  char *swapped;

  str_len = strlen (str);
  array_len = strlen (one);
  swapped = k_new_s (K_STRING_K_SWAP_STRING_SIZE); /*str_len might be zero*/

  strcpy (swapped, str);

  if (strlen (one) != strlen (two))
  {
    // arrays do not match //
    k_del_s (swapped);
    return 0;
  }

  for (str_pos = 0; str_pos < str_len; str_pos++)
  {
    for (array_pos = 0; array_pos < array_len ; array_pos++)
    {
      if (swapped[str_pos] == one[array_pos])
      {
        swapped[str_pos] = two[array_pos];
      }
    }
  }

  return swapped;
}

//// k_trim
//// Returns the string trimmed of white space
char *k_trim (char *str)
{
  int end, start;
  char *temp;

  end = strlen (str);
  start = 0;

  while ((start < end) && (str[start] <= ' '))
  {
    start++;
  }

  while ((start < end) && (str[end - 1] <= ' '))
  {
    end--;
  }

  if ((start > 0) || (end < (signed) strlen (str)))
  {
    return k_sub (str, start, end);
  }
  else
  {
    temp = k_new_s (strlen (str));
    strcpy (temp, str);
    return temp;
  }
}

//// k_trim
//// Returns the string trimmed of white space; deletes the arg
char *k_trimd (char *str)
{
  int end, start;
  char *temp;

  end = strlen (str);
  start = 0;

  while ((start < end) && (str[start] <= ' '))
  {
    start++;
  }

  while ((start < end) && (str[end - 1] <= ' '))
  {
    end--;
  }

  if ((start > 0) || (end < (signed) strlen (str)))
  {
    temp = k_subd (str, start, end);
    return temp;
  }
  else
  {
    return str;
  }
}

#endif
