/***************************************************************************\
  "parser.hpp" - Parser Module (Phase C)
  Revision 19; 10 Nov 2002 - 15 March 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

#ifndef PARSER_HPP
#define PARSER_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

class Parser
{
  public:
    void parse ();

  private:
    int mxMainCallLocation;
    Token *mToken;    // current token
    TokenCode mCode;  // code of current token
    bool bEndOfBlock; // tells the parser_parse function to ignore the lack
                      //  of an EOL because a block just ended.
    Symbol *mSymNL, *mSymMAIN;

    // Parse Statements //
    void mParseStatement ();

    void mParseLabel ();
    void mParseGoto ();
    void mParseDefinition ();
    void mParseConstDefinition ();
    void mParseIdentifierStatement ();
    void mParse_prim_cin ();
    void mParse_prim_cout ();
    void mParse_prim_rand ();
    void mParse_prim_stop ();
    void mParse_prim_asc ();
    void mParse_prim_fioX ();

    // Parse Blocks //
    void mParseBlock (int anInitialIndent);
    void mParseRoutine ();
    void mParseReturn ();
    void mParseCond ();
    void mParseLoop ();

    // Parse Expressions //
    Type mParseExpression ();             // =, <>, <, >, <=, >=
    Type mParseSimpleExpression ();       // +u, -u, +, -, or
    Type mParseTerm ();                   // *, /, and
    Type mParseFactor ();                 // identifier, literal, paren expression
    Symbol *mParseRoutineCall (Symbol *symbol, bool canOmitParens); // routine call        

    // Special //
    void mEnsureInRoutine (char *illegal_statement);
    void mSkipEOL ();
    void mEnsureSkipEOL ();
    bool mSync (TokenCode aCode, char *aError);
    void mSeekToEOL ();
    void mGetToken ();
    Symbol *mMakeTemp (Type t, char *as = NULL);
    void mPutDeclaration (Symbol *sym, Type t);
};

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

void Parser::parse ()
{
  bEndOfBlock = false;

  // Create the "nl" or new line variable
  mSymNL = gScopeManager->globalEnter ("nl");
  mSymNL->set (repIdent, tStr, true/*const*/, "\n");
  int ten = 10;
  Symbol *sym10 = gScopeManager->globalEnter ("10");
  sym10->set (repLiteral, tInt, true, &ten);
  gICM->put (new Statement (ttDS, mSymNL));
  gICM->put (new Statement (ttPUSH, sym10));
  gICM->put (new Statement (ttASC, NULL));
  gICM->put (new Statement (ttPOP, mSymNL));

  mSymMAIN = NULL;
  gCurrRout = NULL;

  // put a DBGON instruction to turn on high-level language
  //   debugging support in the output file
  if (Global::bDebug)
  {
    gICM->put (new Statement (ttDBGON, NULL));
  }

  // put a 'nop' instruction as a placeholder for when //
  //  the 'main' routine is encoutered                 //
  mxMainCallLocation = gICM->xCurr;
  gICM->put (new Statement (ttNOP, NULL));

  // put a 'stop' instruction to stop the program //
  gICM->put (new Statement (ttSTOP, NULL));

  /////////////
  // PARSING //
  /////////////

  mToken = new Token ();    //
  mToken->code = codeStart; // Passed to lex_get at the start of the file
  mGetToken();              //
  mSkipEOL();

  do {
    // Insert a line number marker //
    gICM->putLine ();

    // Parse a Statement //
    mParseStatement ();

    // Handles empty files //
    if (mCode == codeEOF) break;

    // Unless a block just ended, check for EOL //
    if (bEndOfBlock == true)
      bEndOfBlock = false;
    else
      mEnsureSkipEOL();

  } while (mCode != codeEOF);

  delete mToken;

  if (mSymMAIN == NULL)
  {
    err (true, EF_MAIN_NOT_DEF, NULL);
  }
  else if ( ((Routine*)(mSymMAIN->value))->returnType != tVoid )
  {
    err (true, EF_MAIN_NOT_VOID, gvsTypeStrings[((Routine*)(mSymMAIN->value))->returnType]);
  }

  // put a 'stop' instruction to stop the program //
  gICM->put (new Statement (ttSTOP, NULL));
}

///////////////////
// Special Parse //
///////////////////

void Parser::mParseStatement ()
{
  switch (mCode)
  {
    case (codePOUNDSIGN):
      mParseLabel ();
      break;
    case (codeGOTO):
      mParseGoto ();
      break;  
    case (codeDEF):
      mParseDefinition ();
      break;
    case (codeCONST):
      mParseConstDefinition ();
      break;
    case (codeIdentifier):
      mParseIdentifierStatement ();
      break;
    case (code_PRIM_CIN):
      mParse_prim_cin ();
      break;
    case (code_PRIM_COUT):
      mParse_prim_cout ();
      break;
    case (code_PRIM_RAND):
      mParse_prim_rand ();
      break;
    case (code_PRIM_STOP):
      mParse_prim_stop ();
      break;
    case (code_PRIM_ASC):
      mParse_prim_asc ();
      break;
    case (code_PRIM_FIO_OR):
    case (code_PRIM_FIO_OW):
    case (code_PRIM_FIO_C):
    case (code_PRIM_FIO_EOF):
    case (code_PRIM_FIO_R):
    case (code_PRIM_FIO_W):
      mParse_prim_fioX ();
      break;
    case (codeROUT):
      mParseRoutine ();
      break;
    case (codeRETURN):
      mParseReturn ();
      break;
    case (codeCOND):
      mParseCond ();
      break;
    case (codeLOOP):
      mParseLoop ();
      break;
    case (codeEOL):
    case (codeEOF):
      break;
    default:
      err (true, EF_INVALID_PARSE_POINT, gvsTokenCodeStrings[mCode]);
      errterm ();
      break;
  }
}

///////////////////////
// Statement Parsing //
///////////////////////

void Parser::mParseLabel ()
{
  Symbol *symbolA;

  mEnsureInRoutine ("label"); // trigger error if not in routine

  mGetToken (); // now: identifier

  // Search for the label's identifier and enter it //
  symbolA = gScopeManager->globalSearch  (mToken->lexeme);
  if (symbolA) // label must have been created by mParseGoto
  {
    if (symbolA->type != tLabel)
    {
      err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[symbolA->type]);
    }
  }
  else if (!symbolA) // create label
  {
    symbolA = gScopeManager->globalEnter (mToken->lexeme);
    symbolA->set (repIdent, tLabel, false, NULL);
  }

  gICM->put (new Statement (ttColon, symbolA));

  mGetToken (); // now: EOL or statement

  // The label is directly followed by a statement //
  if (mCode != codeEOL)
  {
    mParseStatement ();
  }
}

void Parser::mParseGoto ()
{
  Symbol *symbolA;

  mEnsureInRoutine ("goto"); // trigger error if not in routine

  mGetToken (); // now: identifier

  // Search for the label's identifier and enter it //
  symbolA = gScopeManager->globalSearch  (mToken->lexeme);
  if (symbolA) // label must have been created already
  {
    if (symbolA->type != tLabel)
    {
      err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[symbolA->type]);
    }
  }
  else if (!symbolA) // create label
  {
    symbolA = gScopeManager->globalEnter (mToken->lexeme);
    symbolA->set (repIdent, tLabel, false, NULL);
  }

  gICM->put (new Statement (ttJMP, symbolA));

  mGetToken (); // now: EOL
}

void Parser::mParseDefinition ()
{
  Symbol *symbolA;
  char *lexemeA;
  Type typeA, typeB;
  bool defineAndAssign;

  while (true)
  {
    mGetToken (); // now: identifier
    if (!mSync (codeIdentifier, EF_MISSING_IDENTIFIER)) return;

    lexemeA = k_clone (mToken->lexeme);

    mGetToken (); // now: ':'
    if (!mSync (codeCOLON, EF_MISSING_TYPE_DEF_OP)) return;

    mGetToken (); // now: type

    // determine the type of the definition //
    if (mCode == codeINT) typeA = tInt;
    else if (mCode == codeBOOL) typeA = tBool;
    else if (mCode == codeRAT) typeA = tRat;
    else if (mCode == codeSTR) typeA = tStr;
    else
    {
      err (true, EF_NOT_VALID_TYPE, mToken->lexeme);
      typeA = tInt; // set as int for recovery
    }

    mGetToken (); // now: '<-' or EOL

    // definition is also an assignment //
    if (mCode == codeARROW)
    {
      mEnsureInRoutine ("assignment"); // trigger error if not in routine

      mGetToken (); // now expression
      typeB = mParseExpression ();

      if (!( ((typeA == tInt) && (typeB == tInt)) ||
             ((typeA == tRat) && (typeB == tRat)) ||
             ((typeA == tStr) && (typeB == tStr)) ||
             ((typeA == tBool) && (typeB == tBool))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
      }

      defineAndAssign = true;
    }
    // definition is not an assignment //
    else
    {
      defineAndAssign = false;
    }

    // Search for the variable's identifier and enter it //
    symbolA = gScopeManager->localSearchOnly  (lexemeA);
    if (symbolA)
    {
      // The symbol has already been defined, cause error //
      err (true, EF_ALREADY_DEFINED, lexemeA);
    }
    else if (!symbolA)
    {
      // The symbol is created //
      symbolA = gScopeManager->localEnter (lexemeA);
      symbolA->set (repIdent, typeA, false, NULL);
    }

    // Insert the declaration operation //
    mPutDeclaration (symbolA, typeA);

    // See if the variable is assigned a value from the statement.
    //   If it has not, assign the variable a zero-initialized
    //   value.
    if (defineAndAssign) // define and assign
    {
      gICM->put (new Statement (ttPOP, symbolA));
    }
    else // define only
    {
      /* THIS CAN BE OMITTED BECAUSE PORTIA AUTO-INITIALIZES TO ZERO
      Symbol *symbolB = new Symbol ("assignment");
      symbolB->set (repLiteral, symbolA->type, true, NULL);
      gScopeManager->addSymbolToList (symbolB);

      gICM->put (new Statement (ttPUSH, symbolB));
      gICM->put (new Statement (ttPOP, symbolA));
      */
    }

    // Add the variable to the routine's local variable list //
    if (gCurrRout != NULL)
      gCurrRout->addLocalVar (symbolA);

    k_del_s (lexemeA);

    if (mCode != codeCOMMA) // there is not another variable defined
    {
      break;
    }
  }
}

void Parser::mParseConstDefinition ()
{
  Symbol *symbolA;
  char *lexemeA;
  Type typeA, typeB;

  mEnsureInRoutine ("const definition");

  while (true)
  {
    mGetToken (); // now: identifier
    if (!mSync (codeIdentifier, EF_MISSING_IDENTIFIER)) return;

    lexemeA = k_clone (mToken->lexeme);

    mGetToken (); // now: ':'
    if (!mSync (codeCOLON, EF_MISSING_TYPE_DEF_OP)) return;

    mGetToken (); // now: type

    // determine the type of the definition //
    if (mCode == codeINT) typeA = tInt;
    else if (mCode == codeBOOL) typeA = tBool;
    else if (mCode == codeRAT) typeA = tRat;
    else if (mCode == codeSTR) typeA = tStr;
    else
    {
      err (true, EF_NOT_VALID_TYPE, mToken->lexeme);
      typeA = tInt; // set as int for recovery
    }

    mGetToken (); // now: '<-'
    if (!mSync (codeARROW, EF_MISSING_EXPRESSION)) return;


    mGetToken (); // now expression
    typeB = mParseExpression ();

      if (!( ((typeA == tInt) && (typeB == tInt)) ||
             ((typeA == tRat) && (typeB == tRat)) ||
             ((typeA == tStr) && (typeB == tStr)) ||
             ((typeA == tBool) && (typeB == tBool))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
      }

    // Search for the variable's identifier and enter it //
    symbolA = gScopeManager->localSearchOnly  (lexemeA);
    if (symbolA)
    {
      // The symbol has already been defined, cause error //
      err (true, EF_ALREADY_DEFINED, lexemeA);
    }
    else if (!symbolA)
    {
      // The symbol is created //
      symbolA = gScopeManager->localEnter (lexemeA);
      symbolA->set (repIdent, typeA, true, NULL);
    }

    // Add the variable to the routine's local variable list //
    if (gCurrRout != NULL)
      gCurrRout->addLocalVar (symbolA);

    // Insert the declaration operation //
    mPutDeclaration (symbolA, typeA);

    // Insert intermediate code //
    gICM->put (new Statement (ttPOP, symbolA));

    k_del_s (lexemeA);

    if (mCode != codeCOMMA) // there is not another const defined
    {
      break;
    }
  }
}

void Parser::mParseIdentifierStatement ()
{
  Symbol *symbolA;
  Type typeB;

  // Search for the identifier and enter it if necessary //
  symbolA = gScopeManager->localSearch (mToken->lexeme);
  if (!symbolA)
  {
    int ntemp;

    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);

    // Enter the symbol as int for recovery //
    symbolA = gScopeManager->localEnter (mToken->lexeme);
    ntemp = 0;
    symbolA->set (repIdent, tInt, false, &ntemp);
  }

  // Statement is a Routine call //
  if (symbolA->type == tRoutine)
  {
    Symbol *temp;

    mEnsureInRoutine ("routine call"); // trigger error if not in routine  

    if (((Routine*)(symbolA->value))->returnType != tVoid)
    {
      err (true, EF_RETURN_VAL_NOT_ASSIGNED, NULL);
    }

    temp = mParseRoutineCall (symbolA, true); // now: EOL
  }

  // Statement is an Assignment //
  else
  {
    mEnsureInRoutine ("assignment"); // trigger error if not in routine

    // Ensure no assignment to constant //
    if (symbolA->constant == true)
    {
      err (true, EF_CONST_ASSIGNMENT, mToken->lexeme);
    }

    // :
    mGetToken ();
    if (!mSync (codeARROW, EF_MISSING_ASSIGNMENT_OP)) return;

    // Expression //
    mGetToken ();
    typeB = mParseExpression ();

    if (!((((symbolA->type == tInt) || (symbolA->type == tRat)) &&
           ((typeB == tInt) || (typeB == tRat))) ||
          (((symbolA->type == tStr) && (typeB == tStr))) ||
          (((symbolA->type == tBool) && (typeB == tBool)))
         )
       )
    {
      err (true, EF_TYPE_INCOMPATIBLE, NULL);
    }

    gICM->put (new Statement (ttPOP, symbolA));
  }
}

void Parser::mParse_prim_cin ()
{
  Symbol *symbol;

  mEnsureInRoutine ("_prim_cin"); // trigger error if not in routine

  mGetToken (); // now: variable

  // Get token's symbol //
  symbol = gScopeManager->localSearch (mToken->lexeme);
  if (!symbol)
  {
    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);

    // Enter the symbol as str for recovery //
    symbol = gScopeManager->localEnter (mToken->lexeme);
    symbol->set (repIdent, tStr, false, NULL);
  }

  // Ensure no assignment to constant //
  if (symbol->constant == true)
  {
    err (true, EF_CONST_ASSIGNMENT, mToken->lexeme);
  }

  gICM->put (new Statement (ttCIN, NULL));

  // Convert from string to symbol's type //
  switch (symbol->type)
  {
    case tBool:
      gICM->put (new Statement (ttTOB, NULL));
      break;
    case tInt:
      gICM->put (new Statement (ttTOI, NULL));
      break;
    case tRat:
      gICM->put (new Statement (ttTOR, NULL));
      break;
    case tStr:
      gICM->put (new Statement (ttTOS, NULL));
      break;
  }

  gICM->put (new Statement (ttPOP, symbol));

  mGetToken (); // now: EOL
}

void Parser::mParse_prim_cout ()
{
  mEnsureInRoutine ("_prim_cout"); // trigger error if not in routine

  mGetToken (); // now first token of exp or eol

  if (mCode != codeEOL)
  {
    mParseExpression ();

    gICM->put (new Statement (ttCOUT, NULL));
  }
  else
  {
    gICM->put (new Statement (ttPUSH, mSymNL));
    gICM->put (new Statement (ttCOUT, NULL));
  }
}

void Parser::mParse_prim_rand ()
{
  Symbol *symbol;

  mEnsureInRoutine ("_prim_rand"); // trigger error if not in routine

  mGetToken (); // now: variable

  // Get token's symbol //
  symbol = gScopeManager->localSearch (mToken->lexeme);
  if (!symbol)
  {
    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);

    // Enter the symbol as str for recovery //
    symbol = gScopeManager->localEnter (mToken->lexeme);
    symbol->set (repIdent, tStr, false, NULL);
  }

  // Ensure no assignment to constant //
  if (symbol->constant == true)
  {
    err (true, EF_CONST_ASSIGNMENT, mToken->lexeme);
  }

  if (symbol->type != tInt)
  {
    err (true, EF_TYPE_INCOMPATIBLE, NULL);
  }

  gICM->put (new Statement (ttRAND, NULL));
  gICM->put (new Statement (ttPOP, symbol));

  mGetToken (); // now: EOL
}

void Parser::mParse_prim_stop ()
{
  mEnsureInRoutine ("_prim_stop"); // trigger error if not in routine

  gICM->put (new Statement (ttSTOP, NULL));
  mGetToken (); // now: EOL
}

//// Syntax: _prim_asc <target:str> <expression:int>
void Parser::mParse_prim_asc ()
{
  Symbol *symbolA;
  Type typeB;

  mEnsureInRoutine ("_prim_asc"); // trigger error if not in routine

  mGetToken (); // now: target variable

  // Get token's symbol //
  symbolA = gScopeManager->localSearch (mToken->lexeme);
  if (!symbolA)
  {
    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);

    // Enter the symbol as str for recovery //
    symbolA = gScopeManager->localEnter (mToken->lexeme);
    symbolA->set (repIdent, tStr, false, NULL);
  }

  mGetToken (); // now: integer expression

  typeB = mParseExpression (); // now: EOL

  gICM->put (new Statement (ttASC, NULL));
  gICM->put (new Statement (ttPOP, symbolA));
}

//// Syntax: COMMAND <(A)var:int> <(B)var:str>
  //
  // COMMAND         OPERATION     A (int)       B (str)
  // -------------   -----------   -----------   ------------
  // _prim_fio_or    open input    dest handle   filename
  // _prim_fio_ow    open output   dest handle   filename
  // _prim_fio_c     close         handle        ""
  // _prim_fio_r     read          handle        dest line
  // _prim_fio_w     write         handle        line
  // _prim_fio_eof   eof           handle        dest reached (bool)
void Parser::mParse_prim_fioX ()
{
  Symbol *symbolA;
  Symbol *symbolB;

  mEnsureInRoutine (mToken->lexeme); // trigger error if not in routine
  TokenCode command = mCode;

  mGetToken (); // now: A
  if (!mSync (codeIdentifier, EF_MISSING_IDENTIFIER)) return;

  // Get token's symbol //
  symbolA = gScopeManager->localSearch (mToken->lexeme);
  if (!symbolA)
  {
    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);

    // Enter the symbol as str for recovery //
    symbolA = gScopeManager->localEnter (mToken->lexeme);
    symbolA->set (repIdent, tInt, false, NULL);
  }

  mGetToken (); // now: B
  if (!mSync (codeIdentifier, EF_MISSING_IDENTIFIER)) return;

  // Get token's symbol //
  symbolB = gScopeManager->localSearch (mToken->lexeme);
  if (!symbolB)
  {
    // The symbol has not been declared, cause error //
    err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);

    // Enter the symbol as str for recovery //
    symbolB = gScopeManager->localEnter (mToken->lexeme);
    symbolB->set (repIdent, tStr, false, NULL);
  }
  mGetToken (); // now: EOL

  switch (command)
  {
    case (code_PRIM_FIO_OR):
      gICM->put (new Statement (ttFIO_OR, symbolB));
      gICM->put (new Statement (ttPOP, symbolA));
      break;
    case (code_PRIM_FIO_OW):
      gICM->put (new Statement (ttFIO_OW, symbolB));
      gICM->put (new Statement (ttPOP, symbolA));
      break;
    case (code_PRIM_FIO_C):
      gICM->put (new Statement (ttFIO_C, symbolA));
      break;
    case (code_PRIM_FIO_EOF):
      gICM->put (new Statement (ttFIO_EOF, symbolA));
      gICM->put (new Statement (ttPOP, symbolB));
      break;
    case (code_PRIM_FIO_R):
      gICM->put (new Statement (ttFIO_R, symbolA));
      gICM->put (new Statement (ttPOP, symbolB));
      break;
    case (code_PRIM_FIO_W):
      gICM->put (new Statement (ttPUSH, symbolB));
      gICM->put (new Statement (ttFIO_W, symbolA));
      break;
  }  
}

///////////////////
// Block Parsing //
///////////////////

void Parser::mParseBlock (int anInitialIndent)
{
  int i = 0;

  while ((mCode != codeEOF) && (Global::nIndent > anInitialIndent))
  {
    // Insert a line number marker //
    gICM->putLine ();

    // Parse a Statement //
    mParseStatement ();

    // Unless a block just ended, check for EOL //
    if (bEndOfBlock == true)
      bEndOfBlock = false;
    else
      mEnsureSkipEOL();

    i++;
  }    

  if (i == 0)
  {
    err (true, EF_EMPTY_BLOCK, NULL);
  }

  bEndOfBlock = true;
}

void Parser::mParseRoutine ()
{
  Symbol *symRoutine;
  Symbol *symTemp;
  int nInitialIndent;
  Type returnType;
  char *argumentLexemes[GLOBAL_ROUTINE_MAX_PARAMETERS];
  Type argumentTypes[GLOBAL_ROUTINE_MAX_PARAMETERS];
  int iArgument;
  int i;

  nInitialIndent = Global::nIndent;
  iArgument = 0;

  mGetToken (); // now: identifier

  // Search for the routine's identifier and enter it //
  symRoutine = gScopeManager->globalSearch  (mToken->lexeme);
  if (!symRoutine)
  {
    symRoutine = gScopeManager->globalEnter (mToken->lexeme);

    // If this is the 'main' routine, insert the code to execute it //
    if (strcmp (mToken->lexeme, "main") == 0)
    {
      mSymMAIN = symRoutine;

      // delete the 'nop' and set the first instruction //
      gICM->replace (mxMainCallLocation, new Statement (ttCALL, mSymMAIN));
    }
  }
  else
  {
    err (true, EF_ROUT_REDEF, mToken->lexeme);
    errterm ();
  }

  mGetToken (); // now: ':' or '(' or EOL

  // Routine has a non-void type //
  if (mCode == codeCOLON)
  {
    mGetToken (); // now: type

    // determine the type of the definition //
    if (mCode == codeINT) returnType = tInt;
    else if (mCode == codeBOOL) returnType = tBool;
    else if (mCode == codeRAT) returnType = tRat;
    else if (mCode == codeSTR) returnType = tStr;
    else
    {
      err (true, EF_NOT_VALID_TYPE, mToken->lexeme);
      returnType = tInt; // set as int for recovery
    }

    mGetToken (); // now: EOL
  }
  // Routine has a void type //
  else
  {
    returnType = tVoid;
  }

  // Set the Routine's Symbol and the Routine //
  symRoutine->set (repLiteral, tRoutine, false, NULL);
  ((Routine *)(symRoutine->value))->returnType = returnType;

  // Set the position symbol //
  Symbol *symbolRoutineStartLabel = mMakeTemp (tLabel, symRoutine->key);
  gICM->put (new Statement (ttColon, symbolRoutineStartLabel));
  ((Routine *)(symRoutine->value))->location = symbolRoutineStartLabel;

  // Routine may have parameters //
  if (mCode == codeLEFTPAREN)
  {
    mGetToken (); // now: identifier or ')'

    // Loop to parse parameters //
    while (mCode != codeRIGHTPAREN)
    {
      argumentLexemes[iArgument] = k_clone (mToken->lexeme);

      mGetToken (); // now: ':'
      if (!mSync (codeCOLON, EF_MISSING_TYPE_DEF_OP)) return;

      mGetToken (); // now: type

      // determine the type of the definition //
      if (mCode == codeINT) argumentTypes[iArgument] = tInt;
      else if (mCode == codeBOOL) argumentTypes[iArgument] = tBool;
      else if (mCode == codeRAT) argumentTypes[iArgument] = tRat;
      else if (mCode == codeSTR) argumentTypes[iArgument] = tStr;
      else
      {
        err (true, EF_NOT_VALID_TYPE, mToken->lexeme);
        argumentTypes[iArgument] = tInt; // set as int for recovery
      }

      // add the parameter to the routine's parameter list //
      ((Routine *)(symRoutine->value))->addParameter(argumentTypes[iArgument]);

      mGetToken (); // now: ',' or ')'

      if (mCode == codeCOMMA) // there is another variable defined
      {
        mGetToken (); // now: identifier or ')'
      }

      iArgument++;
    }

    mGetToken (); // now: EOL
  }

  mEnsureSkipEOL(); // now: first token of block

  // Loop to define the arguments //
  for (i = 0; i < iArgument; i++)
  {
    Symbol *symCurrentArgument;

    // Search for the variable's identifier and enter it //
    symCurrentArgument = gScopeManager->localSearchOnly  (argumentLexemes[i]);
    if (symCurrentArgument)
    {
      // The symbol has already been defined, cause error //
      err (true, EF_ALREADY_DEFINED, argumentLexemes[i]);
    }
    else if (!symCurrentArgument)
    {
      // The symbol is created //
      symCurrentArgument = gScopeManager->localEnter (argumentLexemes[i]);
      symCurrentArgument->set (repIdent, argumentTypes[i], false, NULL);
      mPutDeclaration (symCurrentArgument, argumentTypes[i]);
    }

    // Add the parameter to the routine's local variable list //
    ((Routine *)(symRoutine->value))->addLocalVar (symCurrentArgument);

    // Insert the instruction //
    gICM->put (new Statement (ttPOPP, symCurrentArgument));

    k_del_s (argumentLexemes[i]);
  }

  // Parse a block of statements //
  gCurrRout = (Routine*)(symRoutine->value);
  mParseBlock (nInitialIndent);

  // Make sure that routines with non-void return types put //
  //  a value on the stack.                                 //
  if (gCurrRout->returnType != tVoid)
  {
    if (gCurrRout->returnedFrom == false)
    {
      err (true, EF_RETURN_NEVER_MADE, symRoutine->key);
    }

    symTemp = mMakeTemp (((Routine*)(symRoutine->value))->returnType,
                         "routine_error_exit");
    gICM->put (new Statement (ttPUSHP, symTemp));
  }

  // Put a 'ret' in the icode in case one was never placed in //
  //  the body of the routine.                                //
  gICM->put (new Statement (ttRET, NULL));

  // Let the rest of the parser know that it is not in a routine //
  gCurrRout = NULL;
}

void Parser::mParseReturn ()
{
  Symbol *symbol;
  Type type;

  mEnsureInRoutine ("return"); // trigger error if not in routine

  mGetToken (); // now first token of exp or EOL

  // Routine does not have type void; push value to stack //
  if (gCurrRout->returnType != tVoid)
  {
    // push return value to stack //
    type = mParseExpression ();

    gCurrRout->returnedFrom = true;

    // expression type does not match routine return type //
    if (type != gCurrRout->returnType)
    {
      err (true, EF_NON_MATCHING_RET_TYPE, gvsTypeStrings[gCurrRout->returnType]);
    }

    symbol = mMakeTemp (type, "return_temp");
    gICM->put (new Statement (ttPOP, symbol));
    gICM->put (new Statement (ttPUSHP, symbol));
  }
  else
  {
    if (mCode != codeEOL)
    {
      err (true, EF_NO_RETURN_FOR_VOID_ROUT, NULL);
      mSeekToEOL ();
    }
  }

  gICM->put (new Statement (ttRET, NULL));
}

void Parser::mParseCond ()
{
  Statement *statementJump;
  Symbol *symbolFalseLabel;
  Type typeA;

  int nInitialIndent = Global::nIndent;

  mGetToken (); // now first token of expression
  typeA = mParseExpression ();

  statementJump = new Statement (ttJF, NULL);
  gICM->put (statementJump);

  if (typeA != tBool)
  {
    err (true, EF_TYPE_INCOMPATIBLE, "\"cond\" expects boolean");
  }

  mEnsureSkipEOL();

  // parse a block of statements //
  mParseBlock (nInitialIndent);

  // Set the false position symbol //
  symbolFalseLabel = mMakeTemp (tLabel, "cond_false_pos");
  gICM->put (new Statement (ttColon, symbolFalseLabel));

  statementJump->sym = symbolFalseLabel;
}

void Parser::mParseLoop ()
{
  Statement *statementJump;
  Symbol *symbolFalseLabel;
  Symbol *symbolStart;
  Type typeA;

  int nInitialIndent = Global::nIndent;

  // Set the start of the loop symbol //
  symbolStart = mMakeTemp (tLabel, "loop_start_pos");
  gICM->put (new Statement (ttColon, symbolStart));

  mGetToken (); // now first token of expression
  typeA = mParseExpression ();

  statementJump = new Statement (ttJF, NULL);
  gICM->put (statementJump);

  if (typeA != tBool)
  {
    err (true, EF_TYPE_INCOMPATIBLE, "\"cond\" expects boolean");
  }

  mEnsureSkipEOL();

  // parse a block of statements //
  mParseBlock (nInitialIndent);

  gICM->put (new Statement (ttJMP, symbolStart));

  // Set the false position symbol //
  symbolFalseLabel = mMakeTemp (tLabel, "loop_false_pos");
  gICM->put (new Statement (ttColon, symbolFalseLabel));

  statementJump->sym = symbolFalseLabel;
}

////////////////////////
// Expression Parsing //
////////////////////////

//// Duties: relational ops
Type Parser::mParseExpression ()
{
  Type typeA;
  Type typeB;
  Type typeC;
  OpCode op;

  // Parse a simple expression //
  typeB = mParseSimpleExpression ();

  // If there is a relational operator, parse another simple expression //
  while ((mCode == codeEQUALS) || (mCode == codeNOTEQUAL) ||
         (mCode == codeLT) || (mCode == codeGT) ||
         (mCode == codeLTEQUAL) || (mCode == codeGTEQUAL)
        )
  {
    if (mCode == codeEQUALS) op = ttEQ;
    else if (mCode == codeNOTEQUAL) op = ttNEQ;
    else if (mCode == codeLT) op = ttLT;
    else if (mCode == codeGT) op = ttGT;
    else if (mCode == codeLTEQUAL) op = ttLTEQ;
    else if (mCode == codeGTEQUAL) op = ttGTEQ;

    // ensure that strings are not used with anything but '=' and '<>' //
    if ((op != ttEQ) && (op != ttNEQ) && (typeB == tStr))
    {
      err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[typeB]);
    }

    mGetToken ();
    typeC = mParseSimpleExpression();

    // ensure that the types match //
    if (!((((typeB == tInt) || (typeB == tRat)) &&
           ((typeC == tInt) || (typeC == tRat))) ||
          (((typeB == tStr) && (typeC == tStr))) ||
          (((typeB == tBool) && (typeC == tBool)))
         )
       )
    {
      err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[typeC]);
    }

    typeA = tBool;

    gICM->put (new Statement (op, NULL));
    typeB = typeA; // set the output of this loop as the input for next
  }

  return typeB; // return the simple expression
}

//// Duties: binary +, -, &, 'or',
Type Parser::mParseSimpleExpression ()
{
  Type typeA;
  Type typeB;
  Type typeC;
  OpCode op;

  // Parse the first term //
  typeB = mParseTerm ();

  // Loop to parse subsequent additive operators and terms //
  while ((mCode == codeMINUS) || (mCode == codePLUS) ||
         (mCode == codeOR) || (mCode == codeANDSIGN))
  {
    if (mCode == codeMINUS) op = ttSUBI; // assume integer output at first
    else if (mCode == codePLUS) op = ttADDI; // assume integer output at first
    else if (mCode == codeOR) op = ttOR;
    else if (mCode == codeANDSIGN) op = ttCAT;

    mGetToken ();
    typeC = mParseTerm();

    if ((op == ttSUBI) || (op == ttADDI))
    {
      // ensure that types match //
      if (!(((typeB == tInt) || (typeB == tRat)) &&
            ((typeC == tInt) || (typeC == tRat))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
      }

      // This section ensures that if either argument is a rat, //
      //  the resulting value will also be a rat                //
      if ((typeB == tRat) || (typeC == tRat))
      {
        typeA = tRat;
        if (op == ttSUBI)      op = ttSUBR;
        else if (op == ttADDI) op = ttADDR;
      }
      else
      {
        typeA = tInt;
      }
    }
    else if (op == ttOR)
    {
      // ensure that types match //
      if (! ((typeB == tBool) || (typeB == tBool)) )
      {
        err (true, EF_TYPE_INCOMPATIBLE, "expected bool");
      }

      typeA = tBool;
    }
    else if (op == ttCAT)
    {
      // ensure that types match //
      if (!( (typeB == tStr) && (typeC == tStr)
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, "expected string");
      }

      typeA = tStr;
    }

    gICM->put (new Statement (op, NULL));
    typeB = typeA; // set the output of this loop as the input for next
  }

  return typeB; // return the term
}

//// Duties: binary *, /, 'and', ^, 'mod'
Type Parser::mParseTerm ()
{
  Type typeA;
  Type typeB;
  Type typeC;
  OpCode op;

  // Parse the first factor //
  typeB = mParseFactor();

  // Loop to parse subsequent multiplicative operators and factors //
  while ((mCode == codeTIMES) || (mCode == codeFORWARDSLASH) ||
         (mCode == codeAND) || (mCode == codeEXPONENT) ||
         (mCode == codeMOD)
        )
  {
    if (mCode == codeTIMES) op = ttMULI;
    else if (mCode == codeFORWARDSLASH) op = ttDIVI;
    else if (mCode == codeAND) op = ttAND;
    else if (mCode == codeEXPONENT) op = ttPOWI;
    else if (mCode == codeMOD) op = ttMODI;

    mGetToken ();
    typeC = mParseFactor();

    if ((op == ttMULI) || (op == ttDIVI))
    {
      // ensure that types match //
      if (!(((typeB == tInt) || (typeB == tRat)) &&
            ((typeC == tInt) || (typeC == tRat))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
      }

      // This section ensures that if either argument is a rat, //
      //  the resulting value will also be a rat                //
      if ((typeB == tRat) || (typeC == tRat))
      {
        typeA = tRat;
        if (op == ttMULI)      op = ttMULR;
        else if (op == ttDIVI) op = ttDIVR;
      }
      else
      {
        typeA = tInt;
      }
    }
    else if (op == ttAND)
    {
      // ensure that types match //
      if (! ((typeB == tBool) || (typeB == tBool)) )
      {
        err (true, EF_TYPE_INCOMPATIBLE, "expected bool");
      }

      typeA = tBool;
    }
    else if (op == ttPOWI)
    {
      // ensure that types match //
      if (!(((typeB == tInt) || (typeB == tRat)) &&
            ((typeC == tInt) || (typeC == tRat))
           )
         )
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
      }

      // This section ensures that if either argument is a rat, //
      //  the resulting value will also be a rat                //
      if ((typeB == tRat) || (typeC == tRat))
      {
        typeA = tRat;
        op = ttPOWR;
      }
      else
      {
        typeA = tInt;
      }
    }
    else if (op == ttMODI)
    {
      // ensure that types match //
      if (!((typeB == tInt) && (typeC == tInt)))
      {
        err (true, EF_TYPE_INCOMPATIBLE, NULL);
      }

      typeA = tInt;
    }

    gICM->put (new Statement (op, NULL));
    typeB = typeA;  // set the output of this loop as the input for next
  }

  return typeB; // return the factor
}

//// Duties: identifier, literal, parenthesized expression, unary negative, unary type conversion
Type Parser::mParseFactor ()
{
  Symbol *symbolA;
  Type typeA, typeB;

  switch (mCode)
  {
    case codeIdentifier:
    {
      // Search for the identifier //
      symbolA = gScopeManager->localSearch (mToken->lexeme);
      if (!symbolA)
      {
        // Not found; Undefined identifier error; Enter it in symbol table //
        err (true, EF_UNDECLARED_IDENTIFIER, mToken->lexeme);

        symbolA = gScopeManager->localEnter (mToken->lexeme);
      }

      // identifier is a routine call //
      if (symbolA->type == tRoutine)
      {
        if (((Routine*)(symbolA->value))->returnType == tVoid)
        {
          err (true, EF_NO_RETURN_VALUE, symbolA->key);
          errterm ();
        }

        symbolA = mParseRoutineCall (symbolA, false); // now: EOL
      }
      else
      {
        mGetToken (); // now: EOL
      }

      typeA = symbolA->type;
      gICM->put (new Statement (ttPUSH, symbolA));
      return typeA;
      break;
    }
    case codeINT:
    case codeRAT:
    case codeBOOL:
    case codeSTR:
    {
      if (mCode == codeINT) typeA = tInt;
      else if (mCode == codeBOOL) typeA = tBool;
      else if (mCode == codeRAT) typeA = tRat;
      else if (mCode == codeSTR) typeA = tStr;

      mGetToken ();
      typeB = mParseFactor ();

      // see if the types are the same //
      if (typeA == typeB)
      {
        err (true, EF_UNNECESSARY_TYPE_CONV, gvsTypeStrings[typeA]);

        return typeB;
      }

      if (typeA == tInt)
        gICM->put (new Statement (ttTOI, NULL));
      else if (typeA == tRat)
        gICM->put (new Statement (ttTOR, NULL));
      else if (typeA == tBool)
        gICM->put (new Statement (ttTOB, NULL));
      else if (typeA == tStr)
        gICM->put (new Statement (ttTOS, NULL));

      return typeA;
      break;
    }
    case codeIntLiteral:
    case codeBoolLiteral:
    case codeRatLiteral:
    case codeStrLiteral:
    {
      if (mCode == codeIntLiteral) typeA = tInt;
      else if (mCode == codeBoolLiteral) typeA = tBool;
      else if (mCode == codeRatLiteral) typeA = tRat;
      else if (mCode == codeStrLiteral) typeA = tStr;

      // Search for it and enter it in the symbol table //
      symbolA = gScopeManager->globalSearch (mToken->lexeme);
      if (!symbolA)
      {
        symbolA = gScopeManager->globalEnter (mToken->lexeme);
        symbolA->set (repLiteral, typeA, true, mToken->value);
      }

      mGetToken ();
      gICM->put (new Statement (ttPUSH, symbolA));
      return typeA;
      break;
    }
    case codeMINUS:
    {
      mGetToken ();
      typeB = mParseFactor ();

      // ensure that only numbers are used //
      if (! ((typeB == tInt) || (typeB == tRat)) )
      {
        err (true, EF_TYPE_INCOMPATIBLE, gvsTypeStrings[typeB]);
      }

      typeA = typeB;

      gICM->put (new Statement (ttNEG, NULL));
      return typeA;
      break;
    }
    case codeLEFTPAREN:
    {
      // Parenthesized subexpression; Call mParseExpression recursively //
      mGetToken ();
      typeA = mParseExpression ();

      // Check for the closing right parenthesis //
      if (mCode == codeRIGHTPAREN)
      {
        mGetToken ();
      }
      else
      {
        err (true, EF_MISSING_RIGHT_PAREN, NULL);
      }

      return typeA;
      break;
    }
    default:
    {
      err (true, EF_INVALID_EXPRESSION, NULL);

      // BUGFIX (29 July 2002, build 0103):
      //   this is added so that a NULL pointer is not returned
      symbolA = gScopeManager->globalEnter ("<invalid expression>");
      symbolA->set (repLiteral, tStr, true, "<invalid expression>");

      return symbolA->type;
      break;
    }
  }
}

Symbol *Parser::mParseRoutineCall (Symbol *symbol, bool canOmitParens)
{
  Routine *routine;
  Symbol *temp;
  bool parensAreOmitted = false;

  mGetToken (); // now: '(' or token after routine call

  // Parse any arguments //
  if ((mCode == codeLEFTPAREN) || canOmitParens)
  {
    Type argumentTypes[GLOBAL_ROUTINE_MAX_PARAMETERS];
    int iArgument;

    iArgument = 0;
    routine = (Routine*) symbol->value;

    if (canOmitParens && (mCode != codeLEFTPAREN))
      parensAreOmitted = true;
    else
      mGetToken (); // now: first argument or ')'

    // Loop to parse each argument in the routine call //
    while ((mCode != codeRIGHTPAREN) && (mCode != codeEOL))
    {
      // Store parsed expression in the array //
      argumentTypes[iArgument] = mParseExpression (); // now: ')' or ','

      // There is no parameter for this argument //
      if (routine->parameters[iArgument] == NULL)
      {
        err (true, EF_NO_PARAMETER_FOR_ARGUMENT, NULL);
        errterm ();
      }
      
      // The argument's type does not match parameter's type //
      if (routine->parameters[iArgument]->type != argumentTypes[iArgument])
      {
        err (true, EF_ARG_TYPE_MISMATCH, NULL);
        errterm ();
      }
      
      if (mCode == codeCOMMA)
      {
        mGetToken (); // now: next argument or ')'
      }

      iArgument++;
    }

    // The number of parameters does not match the number of arguments //
    if (iArgument != routine->numOfParameters)
    {
      err (true, EF_ARG_NUM_MISMATCH, NULL);
    }

    // There were argument(s), push the arguments in reverse order //
    for (; iArgument > 0; iArgument--)
    {
      Symbol *sym = mMakeTemp (argumentTypes[iArgument-1], "arg");
      gICM->put (new Statement (ttPOP, sym));
      gICM->put (new Statement (ttPUSHP, sym));
    }

    // Note: this check is to ensure that if the routine call occurs
    //       outside of a routine that the compiler will not crash
    if (gCurrRout)
    {
      // push all local variables //
      gCurrRout->setCurrLocalVarToTail ();
      temp = gCurrRout->getPrevLocalVar ();
      while (temp != NULL)
      {
        gICM->put (new Statement (ttPUSHL, temp));
        temp = gCurrRout->getPrevLocalVar ();
      }

      if (!parensAreOmitted)
      {
        if (!mSync (codeRIGHTPAREN, EF_MISSING_RIGHT_PAREN)) errterm ();
        mGetToken (); // now: EOL
      }
    }
  }
  else
  {
    // push all local variables //
    gCurrRout->setCurrLocalVarToTail ();
    temp = gCurrRout->getPrevLocalVar ();
    while (temp != NULL)
    {
      gICM->put (new Statement (ttPUSHL, temp));
      temp = gCurrRout->getPrevLocalVar ();
    }
  }

  gICM->put (new Statement (ttCALL, symbol));

  // Note: this check is to ensure that if the routine call occurs
  //       outside of a routine that the compiler will not crash
  if (gCurrRout)
  {
    // pop all the pushed local variables //
    gCurrRout->setCurrLocalVarToHead ();
    temp = gCurrRout->getNextLocalVar ();
    while (temp != NULL)
    {
      gICM->put (new Statement (ttPOPL, temp));
      temp = gCurrRout->getNextLocalVar ();
    }
  }

  if (((Routine*)(symbol->value))->returnType != tVoid)
  {
    Symbol *sym = mMakeTemp (((Routine*)(symbol->value))->returnType,
                             "return_val");
    gICM->put (new Statement (ttPOPP, sym));
    return sym;
  }
  else
  {
    return NULL;
  }
}

//////////////
// Special  //
//////////////

//// Used in statement parsing procedures to ensure that the
////   statement is not outside of a routine
//// Created: 0216, 14 Jan 2003
void Parser::mEnsureInRoutine (char *illegal_statement)
{
  if (gCurrRout == NULL)
  {
    err (true, EF_BAD_STATEMENT_OUTSIDE_ROUT, illegal_statement);
  }
}

void Parser::mSkipEOL ()
{
  // Skip over this EOL and any blank lines before parsing the next statement //
  while (mCode == codeEOL)
  {
    mGetToken ();
  }
}

void Parser::mEnsureSkipEOL ()
{
  // Make sure that an end of line has been reached //
  mSync (codeEOL, EF_MISSING_EOL);

  // Skip over this EOL and any blank lines before parsing the next statement //
  while (mCode == codeEOL)
  {
    mGetToken ();
  }
}

//// Make sure that one of the codes has been reached
bool Parser::mSync (TokenCode aCode, char *aError)
{
  if (mCode != aCode) // code did not match
  {
    if ((strcmp (mToken->lexeme, "") == 0) || (mToken->lexeme == NULL))
      err (true, aError, gvsTokenCodeStrings[mToken->code]);
    else
      err (true, aError, mToken->lexeme);      
    mSeekToEOL ();
    return false;
  }
  return true; // code did match
}

void Parser::mSeekToEOL ()
{
  // To allow the parser to continue, skip tokens until an EOL is reached //
  while ((mCode != codeEOL) && (mCode != codeEOF))
  {
    mGetToken ();
  }
}

void Parser::mGetToken ()
{
  mToken = gLexer->get (mToken);
  mCode = mToken->code;
}

Symbol *Parser::mMakeTemp (Type t, char *as)
{
  char *s;
  
  if (as)
  {
    s = gUnique->make (as);
  }
  else
  {
    s = gUnique->make ("temp");
  }

  Symbol *symbol = new Symbol (s);
  k_del_s (s);

  symbol->set (repIdent, t, false, NULL);
  gScopeManager->addSymbolToList (symbol);

  // Insert the declaration operation //
  mPutDeclaration (symbol, t);

  return symbol;
}

void Parser::mPutDeclaration (Symbol *sym, Type t)
{
  switch (t)
  {
    case tInt:
      gICM->put (new Statement (ttDI, sym));
      break;
    case tBool:
      gICM->put (new Statement (ttDB, sym));
      break;
    case tRat:
      gICM->put (new Statement (ttDR, sym));
      break;
    case tStr:
      gICM->put (new Statement (ttDS, sym));
      break;
  }
}

#endif
