/***************************************************************************\
  "scopemanager.hpp" - Scope Stack Manager Module (Phase C)
  Revision 80; 10 Nov 2002 - 15 March 2004
  Copyright (C) 2002-2004  Kevin P. Albrecht
\***************************************************************************/

#ifndef SCOPEMANAGER_HPP
#define SCOPEMANAGER_HPP

/////////////////////////
// DEFINES //////////////
/////////////////////////

/////////////////////////
// INCLUDES /////////////
/////////////////////////

// Standard //

// Local //
#include "table.hpp"

/////////////////////////
// DECLARATIONS /////////
/////////////////////////

// External //
class ScopeManager
{
  public:
    ScopeManager ();
    ~ScopeManager ();

    void begin ();
    void end ();

    Symbol *localEnter (char *aKey);
    Symbol *localSearch (char *aKey);
    Symbol *localSearchOnly (char *aKey);
    Symbol *globalEnter (char *aKey);
    Symbol *globalSearch (char *aKey);
    void addSymbolToList (Symbol *symbol);

  private:
    Table *mScopeStack[GLOBAL_MAX_SCOPES];
    int mScopeCurr;

    Symbol *mSymbolList[GLOBAL_MAX_SYMBOLS];
    int mSymbolListCurr;
};

/////////////////////////
// DEFINITIONS //////////
/////////////////////////

ScopeManager::ScopeManager ()
{
  int i;

  mScopeCurr = 0;
  mSymbolListCurr = 0;

  // Allocate space for the scope stack //
  for (i = 0; i < GLOBAL_MAX_SCOPES; i++)
  {
    mScopeStack[i] = NULL;
  }

  mScopeStack[mScopeCurr] = new Table ();

  // Allocate space for the symbol list, used to free memory at end of program //
  for (i = 0; i < GLOBAL_MAX_SYMBOLS; i++)
  {
    mSymbolList[i] = NULL;
  }  
}

ScopeManager::~ScopeManager ()
{
  int i;

  delete mScopeStack[0];

  for (i = 0; mSymbolList[i] != NULL; i++)
  {
    delete mSymbolList[i];
  }
}

void ScopeManager::begin ()
{
  mScopeCurr++;

  // Check to see if maximum scopes reached //
  if (mScopeCurr >= GLOBAL_MAX_SCOPES)
  {
    err (true, EF_MAX_SCOPES_REACHED, k_conv_is (GLOBAL_MAX_SCOPES));
    errterm ();
  }

  mScopeStack[mScopeCurr] = new Table ();
}

void ScopeManager::end ()
{
  delete mScopeStack[mScopeCurr];

  mScopeCurr--;

  // Check to see if minimum scopes reached //
  if (mScopeCurr < 0)
  {
    err (true, EF_MIN_SCOPES_REACHED, NULL);
    err (true, EF_INDENT_ERROR, NULL);
    errterm ();
  }
}

Symbol *ScopeManager::localEnter (char *aKey)
{
  Symbol *temp;

  temp = mScopeStack[mScopeCurr]->enter (aKey);
  addSymbolToList (temp);
  temp->uid = gUnique->make ("_");

  return temp;
}

Symbol *ScopeManager::localSearch (char *aKey)
{
  Symbol *temp;
  int i;

  temp = mScopeStack[mScopeCurr]->search (aKey);
  i = mScopeCurr - 1;

  while ((!temp) && (i >= 0))
  {
    temp = mScopeStack[i]->search (aKey);
    i--;
  }

  return temp;
}

Symbol *ScopeManager::localSearchOnly (char *aKey)
{
  return mScopeStack[mScopeCurr]->search (aKey);
}

Symbol *ScopeManager::globalEnter (char *aKey)
{
  Symbol *temp;

  temp = mScopeStack[0]->enter (aKey);
  addSymbolToList (temp);

  return temp;
}

Symbol *ScopeManager::globalSearch (char *aKey)
{
  return mScopeStack[0]->search (aKey);
}

//// This is used to add all symbols to a list to be deleted
////   at program termination; Modules that delete a Symbol
////   should not put it on this list.
void ScopeManager::addSymbolToList (Symbol *symbol)
{
  mSymbolList[mSymbolListCurr] = symbol;
  mSymbolListCurr++;
}

#endif
